package extensions;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.regex.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import dareka.common.*;
import dareka.extensions.Extension;
import dareka.processor.*;

/**
 * @author ListX
 * @rev05(090315)
 */
public class nlGetTag implements Extension, Processor {
	// const
	private static final String GET_TAG_REV = "nlGetTag rev05";
	private static final Pattern GET_LIST_PATTERN = Pattern
			.compile("http://www\\.nicovideo\\.jp/cache/taglist$");
	private static final Pattern ADD_LIST_PATTERN = Pattern
			.compile("http://www\\.nicovideo\\.jp/cache/taglist\\?id=([a-z]{2}\\d+)&tag=(.*?)(&time=\\d+)?$");
	private static final Pattern TAG_XML_PATTERN = Pattern
			.compile("^[a-z]{2}\\d+\\.xml$");

	// var
	private static File tagFolder;
	private static File[] tagFileList;
	private static TreeMap<String, String> tagList = new TreeMap<String, String>();

	// init
	public nlGetTag() throws Exception {
		String rootPath = Config.getString("cacheFolder", "");
		if (rootPath.length() == 0)
			rootPath = "cache";
		tagFolder = new File(rootPath + File.separator + "#tag");
		tagFolder.mkdir();
		if (tagFolder.exists()) {
			Logger.debug(GET_TAG_REV + ": tagCache path=\""
					+ tagFolder.getPath() + '"');
			// rev3܂ł̃^OL^t@Cxml֕ϊ
			File oldTagLog = new File(rootPath + "/#tag.txt");
			if (oldTagLog.exists() && oldTagLog.canRead()) {
				BufferedReader inPut = new BufferedReader(
						new InputStreamReader(new FileInputStream(oldTagLog
								.getPath()), "UTF-8"));
				String buf;
				while ((buf = inPut.readLine()) != null) {
					String data[] = buf.split("\t");
					if (data.length == 2 && data[0] != null
							&& data[0].length() > 0 && data[1] != null
							&& data[0].length() > 0)
						addTagList(data[0], data[1].replaceAll("%", "%25"));
				}
				inPut.close();
				Logger
						.info(GET_TAG_REV
								+ ": "
								+ oldTagLog.getPath()
								+ "(^Oۑpt@C)݂̃Oɕϊ܂BȂϊłƂmFA폜ĂB");
			}
			// ȉ#tagtH_ȉxml̃C|[g
			tagFileList = tagFolder.listFiles();
			for (File tagFile : tagFileList) {
				addTagFromXML(tagFile);
			}
		} else {
			Logger.info(GET_TAG_REV + ": io error! path=\""
					+ tagFolder.getPath() + '"');
		}
	}
	private void addTagFromXML(File tagFile) {
		if (TAG_XML_PATTERN.matcher(tagFile.getName()).matches()) {
			Logger.debug(GET_TAG_REV + ": open " + tagFile.getName());
			String[] ret = getTagByXML(tagFile);
			if (ret == null) {
				Logger.info(GET_TAG_REV + ": " + tagFile.getName()
						+ " is Empty.");
			} else {
				Logger.debug(GET_TAG_REV + ": " + ret[0] + " -> "
						+ ret[1]);
				tagList.put(ret[0], ret[1]);
			}
		} else {
			Logger.info(GET_TAG_REV + ": " + tagFile.getName()
					+ " is not xml.");
		}
	}
	private String[] getTagByXML(File path) {
		try {
			if (path.length() == 0) {
				return null;
			}
			StringBuffer sb = new StringBuffer();
			Document doc = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder().parse(path);
			Element root = doc.getDocumentElement();
			NodeList children = root.getElementsByTagName("tag");
			for (int i = 0, n = children.getLength(); i < n; i++) {
				String buf = children.item(i).getFirstChild().getNodeValue();
				if (buf != null && buf.length() > 0) {
					sb.append(" " + buf);
				}
			}
			String[] ret = { root.getAttribute("id"),
					(sb.length() == 0) ? null : new String(sb.substring(1)) };
			return ret;
		} catch (IOException e) {
			Logger.debug(e);
		} catch (ParserConfigurationException e) {
			Logger.debug(e);
		} catch (SAXException e) {
			Logger.debug(e);
		}
		return null;
	}

	// version
	public String getVersionString() {
		return GET_TAG_REV;
	}

	// Extension interface
	public Object queryInterface(Extension.Type type) {
		if (type == Extension.Type.Extension1)
			return (Extension) this;
		if (type == Extension.Type.Processor1)
			return (Processor) this;
		return null;
	}

	// HTTP̃\bh
	public String[] getSupportedMethods() {
		return new String[] { "GET" };
	}

	// URL̐K\
	public Pattern getSupportedURLAsPattern() {
		return Pattern.compile("^http://www\\.nicovideo\\.jp/cache/taglist.*");
	}

	public String getSupportedURLAsString() {
		return null;
	}

	// Y郊NGXg̍ۂɌĂ΂
	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
			throws IOException {
		String content, url = requestHeader.getURI();
		Matcher match;
		if (GET_LIST_PATTERN.matcher(url).matches()) {
			content = returnTagList();
		} else if ((match = ADD_LIST_PATTERN.matcher(url)).matches()) {
			content = addTagList(match.group(1), match.group(2));
		} else {
			content = "NG";
			Logger.info(GET_TAG_REV + ": NG request\n(url= " + url + ')');
		}
		StringResource r = new StringResource(content);
		r.setResponseHeader("Content-Type", "text/plain");
		return r;
	}

	// ^OXgԂ
	private String returnTagList() {
		String[] tmpList = tagList.keySet().toArray(new String[0]);

		StringBuffer sb = new StringBuffer("tagList={\n");
		for (String id : tmpList) {
			sb.append('"' + id + '"' + ":\"" + tagList.get(id) + "\",\n");
		}
		Logger.info(GET_TAG_REV + ": return tagList");
		return new String(sb).replaceFirst(",?\n$", "\n};").replaceAll("\"",
				"\\\"");// jsƂĕsŝ镶ɑΏ
	}

	// ^OXgɒǉ
	public String addTagList(String key, String data) throws IOException {
		try {
			data = URLDecoder.decode(data, "UTF-8");
			Logger.debug(GET_TAG_REV + ":" + data);
			key = key.replaceAll("[^a-z\\da]|low", "");
			String oldTags = tagList.get(key);
			// DOM\z
			Document doc = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder().newDocument();
			Element root = doc.createElement("tags");
			root.setAttribute("id", key);
			doc.appendChild(root);
			// ^Oǉ
			root.appendChild(doc.createTextNode("\n"));// XMLł̉sp
			String[] buf;
			if (oldTags != null) {
				data = oldTags + " " + data;
			}
			buf = data.replaceAll("(\\s)+", "$1").split(" ");// 2ȏ̋󔒂͏
			TreeMap<String, Boolean> chkTable = new TreeMap<String, Boolean>();// ^OdȂ悤ɂ
			for (String tagStr : buf) {
				if (tagStr != null && tagStr.length() > 0
						&& chkTable.get(tagStr) == null) {
					Logger.debug(GET_TAG_REV + ": " + tagStr + " add");
					chkTable.put(tagStr, true);
					Element tag = doc.createElement("tag");
					tag.appendChild(doc.createTextNode(tagStr));
					root.appendChild(doc.createTextNode("\t"));
					root.appendChild(tag);
					root.appendChild(doc.createTextNode("\n"));// XMLł̉sp
				}
			}
			// XMLƂďo
			File tagFile = new File(tagFolder.getPath() + File.separator + key
					+ ".xml");
			Logger.debug(GET_TAG_REV + ": add tag for " + key + " to "
					+ tagFile.getPath());
			Transformer transformer = TransformerFactory.newInstance()
					.newTransformer();
			transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");// R[hw
			// Cfg̎w
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			FileOutputStream fos = new FileOutputStream(tagFile);
			transformer.transform(new DOMSource(doc), new StreamResult(fos));
			fos.close();
			//XML[hă^Oēo^
			addTagFromXML(tagFile);
			Logger.info(GET_TAG_REV + ": add tag for " + key);
			return "OK";
		} catch (ParserConfigurationException e) {
			Logger.debug(e);
		} catch (TransformerConfigurationException e) {
			Logger.debug(e);
		} catch (TransformerException e) {
			Logger.debug(e);
		}
		return "NG";
	}
}